//-----------------------------------------------------------------------------
// FirstBitGA.cpp
//-----------------------------------------------------------------------------
//*
// An instance of a VERY simple Bitstring Genetic Algorithm
//
//-----------------------------------------------------------------------------

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// standard includes
#include <stdexcept>  // runtime_error 
#include <iostream>   // cout
#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>  // ostrstream, istrstream
#endif

// the general include for eo

#include <eo>
#include <ga.h>

#include "fitness.h"

// Use functions from namespace std
using namespace std;



// GENERAL
//-----------------------------------------------------------------------------
void main_function(int argc, char **argv)
{
// PARAMETRES
  // all parameters are hard-coded!
   unsigned int SEED;      // seed for random number generator
  //  const unsigned int T_SIZE = 3;     // size for tournament selection
   unsigned int VEC_SIZE = SIZE*SIZE;   // Number of bits in genotypes
   unsigned int POP_SIZE;  // Size of population
   unsigned int MAX_GEN;  // Maximum number of generation before STOP
   float CROSS_RATE;      // Crossover rate
   double P_MUT_PER_BIT; // probability of bit-flip mutation
   float MUT_RATE;        // mutation rate

// PARAMETRES
//-----------------------------------------------------------------------------
// instead of having all values of useful parameters as constants, read them:
// either on the command line (--option=value or -o=value)
//     or in a parameter file (same syntax, order independent, 
//                             # = usual comment character 
//     or in the environment (TODO)

  // First define a parser from the command-line arguments
    eoParser parser(argc, argv);

    // For each parameter, define Parameter, read it through the parser,
    // and assign the value to the variable
    eoValueParam<uint32_t> seedParam(42, "seed", "Random number seed", 'S');
    parser.processParam( seedParam );
    SEED = seedParam.value();

   // description of genotype
    //    eoValueParam<unsigned int> vecSizeParam(8, "vecSize", "Genotype size",'V');
    // parser.processParam( vecSizeParam, "Representation" );
    // VEC_SIZE = vecSizeParam.value();

   // parameters for evolution engine
    eoValueParam<unsigned int> popSizeParam(100, "popSize", "Population size",'P');
    parser.processParam( popSizeParam, "Evolution engine" );
    POP_SIZE = popSizeParam.value();

    // init and stop
    eoValueParam<unsigned int> maxGenParam(500, "maxGen", "Maximum number of generations",'G');
    parser.processParam( maxGenParam, "Stopping criterion" );
    MAX_GEN = maxGenParam.value();

   // operators probabilities at the algorithm level
    eoValueParam<double> pCrossParam(0.6, "pCross", "Probability of Crossover", 'C'); 
    parser.processParam( pCrossParam, "Genetic Operators" );
    CROSS_RATE = pCrossParam.value();

    eoValueParam<double> pMutParam(0.1, "pMut", "Probability of Mutation", 'M');
    parser.processParam( pMutParam, "Genetic Operators" );
    MUT_RATE = pMutParam.value();

   // relative rates and private parameters for mutations;
    eoValueParam<double> pMutPerBitParam(0.01, "pMutPerBit", "Probability of flipping 1 bit in bit-flip mutation", 'b');
    parser.processParam( pMutPerBitParam, "Genetic Operators" );
    P_MUT_PER_BIT = pMutPerBitParam.value();

    // name of file to save statistics 
    eoValueParam<string> saveStatsParam("log.log", "LogFile","A file to log statistics",'L');
    parser.processParam( saveStatsParam, "Log Statistics File" );
    string saveStatsName = saveStatsParam.value();


    // the name of the "status" file where all actual parameter values will be saved
    string str_status = parser.ProgramName() + ".status"; // default value
    eoValueParam<string> statusParam(str_status.c_str(), "status","Status file",'S');
    parser.processParam( statusParam, "Persistence" );

   // do the following AFTER ALL PARAMETERS HAVE BEEN PROCESSED
   // i.e. in case you need parameters somewhere else, postpone these
    if (parser.userNeedsHelp())
      {
        parser.printHelp(cout);
        exit(1);
      }
    if (statusParam.value() != "")
      {
	ofstream os(statusParam.value().c_str());
	os << parser;		// and you can use that file as parameter file
      }


// GENERAL
  //////////////////////////
  //  Random seed
  //////////////////////////
  //reproducible random seed: if you don't change SEED above, 
  // you'll aways get the same result, NOT a random run
  rng.reseed(SEED);

// EVAL
  /////////////////////////////
  // Fitness function
  ////////////////////////////
  // Evaluation: from a plain C++ fn to an EvalFunc Object
  eoEvalFuncPtr<Indi> eval(  binary_value );


// INIT
  ////////////////////////////////
  // Initilisation of population
  ////////////////////////////////

  // declare the population
  eoPop<Indi> pop;
  // fill it!
  for (unsigned int igeno=0; igeno<POP_SIZE; igeno++)
    {
      Indi v;           // void individual, to be filled
      for (unsigned ivar=0; ivar<VEC_SIZE; ivar++)
        {
          bool r = rng.flip(); // new value, random in {0,1}
          v.push_back(r);      // append that random value to v
        }
      eval(v);                 // evaluate it
      pop.push_back(v);        // and put it in the population
    }

// OUTPUT
  // sort pop before printing it!
  pop.sort();
  // Print (sorted) intial population (raw printout)
  cout << "Initial Population" << endl;
  cout << pop;

// ENGINE
  /////////////////////////////////////
  // selection and replacement
  ////////////////////////////////////
// SELECT
  // The roulette wheel selection
  eoProportionalSelect<Indi> select;  // T_SIZE in [2,POP_SIZE]
 

// REPLACE

 // the same generational replacement at the moment :-)
  // eoGenerationalReplacement<Indi> replace; 

 // weak  elitism 
  //  eoGenerationalReplacement<Indi> replace_main; 
  //  eoWeakElitistReplacement<Indi> replace(replace_main); 

// OPERATORS
  //////////////////////////////////////
  // The variation operators
  //////////////////////////////////////
// CROSSOVER
  // 1-point crossover for bitstring
  eo1PtBitXover<Indi> xover;
// MUTATION
  // standard bit-flip mutation for bitstring
  eoBitMutation<Indi>  mutation(P_MUT_PER_BIT);

// The operators are  encapsulated into an eoTRansform object
 eoSGATransform<Indi> transform(xover, CROSS_RATE, mutation, MUT_RATE);


// STOP
// CHECKPOINT
  //////////////////////////////////////
  // termination condition
  /////////////////////////////////////
  // stop after MAX_GEN generations

 eoGenContinue<Indi> genCont(MAX_GEN);

 // but now you want to make many different things every generation 
 // (e.g. statistics, plots, ...).
 // the class eoCheckPoint is dedicated to just that:
 // Declare a checkpoint (from a continuator: an eoCheckPoint 
 // IS AN eoContinue and will be called in the loop of all algorithms)
 eoCheckPoint<Indi> checkpoint(genCont);

  // Create a counter parameter
     eoValueParam<unsigned> generationCounter(0, "Gen.");

   // Create an incrementor (sub-class of eoUpdater). Note that the 
     // parameter's value is passed by reference, 
     // so every time the incrementer is updated (every generation),
     // the data in generationCounter will change.
     eoIncrementor<unsigned> increment(generationCounter.value());
  // Add it to the checkpoint, 
     // so the counter is updated (here, incremented) every generation
     checkpoint.add(increment);
  // now some statistics on the population:
     // Best fitness in population
     eoBestFitnessStat<Indi> bestStat;
     // Second moment stats: average and stdev
     eoSecondMomentStats<Indi> SecondStat;
   // Add them to the checkpoint to get them called at the appropriate time
     checkpoint.add(bestStat);
     checkpoint.add(SecondStat);
    
     // A file monitor: will print parameters to ... a File, yes, you got it!
     eoFileMonitor fileMonitor(saveStatsName, " ");

     // the checkpoint mechanism can handle multiple monitors
     checkpoint.add(fileMonitor);
     // the fileMonitor can monitor parameters, too, but you must tell it!
     fileMonitor.add(generationCounter);
     fileMonitor.add(bestStat);
     fileMonitor.add(SecondStat);

   
   
// GENERATION
  /////////////////////////////////////////
  // the algorithm
  ////////////////////////////////////////
  eoSGA<Indi> gga(select, xover, CROSS_RATE, mutation, MUT_RATE, 
                  eval, checkpoint);

  // Apply algo to pop - that's it!
  gga(pop);
  
// OUTPUT
  // Print (sorted) intial population
  pop.sort();
  cout << "FINAL Population\n" << pop << endl;
// GENERAL
}
 // A main that catches the exceptions

int main(int argc, char **argv)
{

    try
    {
        main_function(argc, argv);
    }
    catch(exception& e)
    {
        cout << "Exception: " << e.what() << '\n';
    }

    return 1;
}
